# Caching data for others

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001,2002  Paul Francis Harrison
#    Copyright (C) 2002  Monash University
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# TODO: barf if can't offload
# TODO: barf if no caches in pool

import time, random, threading

from settings import *
from utility import *
from error import Error
import hash

class Cache_item:
    def __init__(self, names,data,expiry_interval,call):
        self.names = names
        self.data = data
        self.expiry_time = time.time() + expiry_interval
        self.call = call

#def cache_task(cache):
#  while 1:
#    cache.lock.acquire()
#    try:
#      if not cache.running:
#        return
#
#      cache.data['up time'] = int(time.time() - cache.start_time)
#
#    finally:
#      cache.lock.release()
#      
#    time.sleep(1)

class Cache(Task_manager):
    def __init__(self, node):
        Task_manager.__init__(self)

        self.node = node

        self.start_time = time.time() # for up_time

        self.cache_pool = [ ]   # list of known caches
        self.cache_pool_last_refresh  = 0.0
        self.cache_pool_original_size = 0
        self.cache_pool_lock = threading.RLock()

        self.data = {
            'type'            : 'service data cache',
            'up time'         : 0,  
            'max expiry time' : 30*24*60*60 
        }

        self.name = hash.hash_of('service data cache')

        self.queued_items = [ ]

        self.cached_items = { }

    def data_function(self):
        up_time = time.time() - self.start_time

        # Try to make it a reasonable number (in case someone's reset the
        # system clock).  We could make a more accurate guess than this, but
        # up-time isn't used for much.
        # Note: this code is ~duplicated in name_server.py.
        if up_time < 0.0:
            up_time = 10.0 * 60
            self.start_time = time.time() - up_time
        elif up_time > 365 * 24 * 60 * 60.0:
            up_time = 25 * 60 * 60.0
            self.start_time = time.time() - up_time
        
        self.data['up time'] = int(up_time)
        return self.data

    def start(self):
        Task_manager.start(self)

        self.node.publish(self.name,self.data_function,cache_redundancy)
        self.node.add_handler('data cache store', self)
        self.node.add_handler('data cache store multi', self)
        self.node.add_handler('data cache remove', self, ('any',))

        #Task(cache_task,self).start()

    def stop(self):
        self.node.unpublish(self.data_function)
        self.node.remove_handler('data cache store')  # lock, name, data, expiry_interval
        self.node.remove_handler('data cache store multi')  # lock, [name,name,...], data, expiry_interval
        self.node.remove_handler('data cache remove') # key

        for item in self.cached_items.items():
            def task(self,item=item):
                self.publish(item[1].names,item[1].data,item[0],1,int(item[1].expiry_time-time.time()))
            Task(task,self).start_no_overload()

        Task_manager.stop(self)

    def handle(self, request,address,call_id):
        check.check_matches(request, (types.StringType,))
        check.check_is_af_inet_address(address)
        
        # Idempotence is a bugger

        if request[0] == 'data cache store' or \
             request[0] == 'data cache store multi':
            self.lock.acquire()
            try:
                if self.cached_items.has_key(request[1]):
                    if self.cached_items[request[1]].call == (address,call_id):
                        return None
                    return Error('already storing')

                if request[0] == 'data cache store':
                    names = [ request[2] ]
                else:
                    names = request[2]

                if len(names) > max_cache_names:
                    return Error('too many names')
                
                item = Cache_item(names,request[3],request[4],(address,call_id))
 
                self.cached_items[request[1]] = item

                for name in names:
                    self.node.publish(name,request[3],cache_redundancy)
            finally:
                self.lock.release()
            return None

        if request[0] == 'data cache remove':
            self.lock.acquire()
            try:
                lock = hash.hash_of(request[1])
                if self.cached_items.has_key(lock):
                    self.node.unpublish(self.cached_items[lock].data)
                    del self.cached_items[lock]
            finally:
                self.lock.release()

            return None

    def refresh_pool(self):
        self.cache_pool_lock.acquire()
        try:
            if self.cache_pool_last_refresh+10*60 > time.time() or \
                 len(self.cache_pool) > self.cache_pool_original_size/2:
                return

            cache_pool = [ ]

            pipe = self.node.retrieve(self.name, cache_redundancy)

            while len(cache_pool) < 50 and not pipe.finished():
                list = pipe.read_all()
                for item in list:
                    check.check_matches(item, ('af_inet_address', 'any'))
                    # Proof: @E6.  Combine with @E10 and the fact that we don't
                    # pass pipe to anything else (nor write to it ourselves
                    # other than through pipe.read_all()) to show that @E6
                    # isn't broken by some other write.
                    
                    if type(item[1]) != types.DictionaryType:
                        # todo: consider printing debug info
                        continue
                    
                    if item[1].get('type','') == 'service data cache':
                        #cache_pool.append((-item[1]['up time'],item[0],item[1]))

                        up_time = item[1].get('up time')

                        if type(up_time) != types.IntType:
                            # todo: consider printing debug info
                            continue
                        
                        # Treat all up for over a day as equal to share load
                        if up_time > 24*60*60:
                            up_time = 24*60*60 + random.random()
                        
                        cache_pool.append((-up_time,item[0],item[1]))
                time.sleep(1)

            pipe.stop()

            self.cache_pool = cache_pool
            self.cache_pool.sort()
            self.cache_pool_original_size = len(self.cache_pool)
        finally:
            self.cache_pool_lock.release()
    
    def publish(self, names,data,lock,redundancy=cache_redundancy,expiry=None):
        """Returns a non-negative integer representing the number of nodes at
             which the data has been stored."""
        self.cache_pool_lock.acquire()
        try:
            self.refresh_pool()

            if not expiry:
                expiry = self.data['max expiry time']

            if expiry < 0:
                return 0

            pool = self.cache_pool[:]
            bad_items = [ ]
        finally:
            self.cache_pool_lock.release()

        pos = 0
        n   = 0
        while n < redundancy:
            if len(pool) <= pos:
                return n

            try:
                if len(names) == 1:
                    self.node.call(pool[pos][1],('data cache store',lock,names[0],data,expiry))
                else:
                    self.node.call(pool[pos][1],('data cache store multi',lock,names,data,expiry))
            except Error, error:
                if error.message <> 'already storing':
                    bad_items.append(pool[pos])
                pos = pos + 1
                continue
            pos = pos + 1
            n   = n   + 1

        self.cache_pool_lock.acquire()
        try:
            for item in bad_items:
                if item in self.cache_pool:
                    self.cache_pool.remove(item)
        finally:
            self.cache_pool_lock.release()
            
        return n


# vim: set expandtab :
