#!/usr/bin/python

# Magic Module

#    Magic - Python module to classify like the 'file' command using a 'magic' file
#    See: 'man 4 magic' and 'man file'
#
#    Copyright (C) 2002 Thomas Mangin
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import re
import string
import convert

# TODO: Need to have a checksum on the cache and source file to update at object creation
# TODO: This program take some input file, we should check the permission on those files ..
# TODO: Some final code cleanup (pycheck complains about code too long ...)
# BUG/LIMITATION: Do not handle all the magic file definition
# BUG: Windows can not read the created cache file (same problem with pickle and cpickle)

class Magic:

    data_size = { 'byte':1, 'short':2, 'long':4, 'string':1, 'pstring':1, 'date': 4, 'ldate': 4 }
    type_size = { 'b':1, 'B':1, 's':2, 'S':2, 'l':4, 'L':5 }


    se_offset_abs="^\(([0\\\][xX][\dA-Fa-f]+|[0\\\][0-7]*|\d+)(\.[bslBSL])*\)"
    se_offset_add="^\(([0\\\][xX][\dA-Fa-f]+|[0\\\][0-7]*|\d+)(\.[bslBSL])*([-+])([0\\\][xX][\dA-Fa-f]+|[0\\\][0-7]*|\d+)\)"
     
    def __split (self,line):
        result = ''
        split = line.split()

        again = 1

        # Make sure the split function did not split too much
        while again:
            again = 0
            pos = 0
            part = []
            top = len(split)
            while pos < top:
                if convert.is_final_dash(split[pos]):
                    result = split[pos] + ' '
                    index = line.find(result)
                    if index != -1:
                        char = line[index+len(result)]
                        if char != ' ' and char != '\t':
                            pos += 1
                            result += split[pos]
                            again = 1
                else:
                    result = split[pos]

                part.append(result)
                pos += 1
            split = part
        
        return part

    def __strip_start (self,char,text):
        if text[0] == char:
            return text[1:]
        return text
    
    def __direct_offset(self,text):
        if text[0] == '(' and text[-1] == ')':
            return 0
        return 1

    def __level(self,text):
        return string.count(text,'>')

    def __endian (self,full_type):
        if full_type.startswith('be'):
            return 'big'
        elif full_type.startswith('le'):
            return 'little'
        return 'local'

    def __kind (self,full_type,endian):
        if endian == 'local':
            kind = full_type
        else:
            kind = full_type[2:]

        # XXX: string case and white space compaction option not implemented
        # XXX: Not very used ...
        if kind.startswith("string/"):
            NOT_DONE_YET=kind[7:]
            kind="string"

        # XXX: No idea what are those extension
        if kind.startswith("ldate-"):
            NOT_DONE_YET=kind[6:]
            kind="ldate"
    
        return kind

    def __oper_mask (self,text):
        type_mask_and = string.split(text,'&')
        type_mask_or = string.split(text,'^')

        if len(type_mask_and) > 1:
            oper = '&'
            mask = convert.convert(type_mask_and[1])
            rest = type_mask_and[0]
            return (oper,mask,rest)
        elif len(type_mask_or) > 1:
            oper = '^'
            mask = convert.convert(type_mask_or[1])
            rest = type_mask_or[0]
            return (oper,mask,rest)
        else:
            return ('',0L,text)
            
    def __test_result (self,test_result):
        if test_result[0] in "=><&!^":
            test   = test_result[0]
            result = test_result[1:]
            return (test,result)
        elif test_result == 'x':
            test = 'x'
            result = 'x'
            return (test,result)
        else:
            test = '='
            result = test_result
            return (test,result)

    def __offset (self,text):
        direct = self.__direct_offset(text)
        offset_type = 'l'
        offset_delta = 0L
        offset_relatif = 0L

        # Get the offset information
        if direct:
            offset_delta = convert.convert(text)
        else:
            match_abs = re.compile(self.se_offset_abs).match(text)
            match_add = re.compile(self.se_offset_add).match(text)

            if match_abs:
                offset_relatif = convert.convert(match_abs.group(1))

                if match_abs.group(2) != None:
                    offset_type = match_abs.group(2)[1]


            elif match_add:
                offset_relatif = convert.convert(match_add.group(1))

                if match_add.group(2) != None:
                    offset_type = match_add.group(2)[1]

                if match_add.group(3) == '-':
                    offset_delta = 0L - match_add.group(4)
                else:
                    offset_delta = convert.convert(match_add.group(4))

        return (direct,offset_type,offset_delta,offset_relatif)
    
    def __mime (self,list):
        mime=''
        for name in list:
            mime += name + " "

        mime = mime.rstrip()

        mime = mime.replace("\\a","\a")
        mime = mime.replace("\\b","\b")
        mime = mime.replace("\\f","\f")
        mime = mime.replace("\\n","\n")
        mime = mime.replace("\\r","\r")
        mime = mime.replace("\\t","\t")
        mime = mime.replace("\\v","\v")

        return mime

    def __data (self,result):
        pos = 0
        data = []

        while pos < len(result):
            if convert.is_c_escape(result[pos:]):
                data.append(result[pos:pos+2])
                pos += 2
            else:
                base = convert.which_base(result[pos:])

                if base == 0:
                    data.append(ord(result[pos])*1L)
                    pos += 1
                else:
                    size_base = convert.size_base(base)
                    size_number = convert.size_number(result[pos:])
                    start = pos + size_base
                    end = pos + size_number
                    nb = convert.base10(result[start:end],base)
                    pos += size_number
                    data.append(nb*1L)
        return data

    def __init__ (self,filename,cachename=""):
        self.entries = 0

        if cachename:
            import os
            
            if not os.path.isfile(cachename):
                self.read_magic(filename)
                self.write_cache(cachename)
        
            self.read_cache(cachename)
        else:
            self.read_magic(filename)
        
    def read_magic (self,magic_file):
        self.magic = []

        try:
            f = open(magic_file,'rt')
        except:
            raise IOError, "No valid magic file called \"" + str(magic_file) + "\""
    
        acc_level = []
        acc_direct = []
        acc_offset_relatif = []
        acc_offset_type = []
        acc_offset_delta = []
        acc_endian = []
        acc_kind = []
        acc_oper = []
        acc_mask = []
        acc_test = []
        acc_data = []
        acc_length = []
        acc_mime = []

        for line in f.readlines():
            line = line.strip()

            if line and not line.startswith('#'):
                part = self.__split(line)
                
                # XXX: ">>10 string" may not be valid so skip it ..
                if len(part) < 4:
                    continue

                # Get the level of the test
                level = self.__level(part[0])

                # XXX: What does the & is used for in ">>&2" as we do not know skip it
                offset_string = self.__strip_start('&',part[0][level:])

                # offset such as (<number>[.[bslBSL]][+-][<number>]) are indirect offset
                (direct,offset_type,offset_delta,offset_relatif) = self.__offset(offset_string)
        
                # The type can be associated to a netmask
                (oper,mask,rest) = self.__oper_mask(part[1])

                # No idea what this 'u' is so skip it
                full_type = self.__strip_start('u',rest)

                endian = self.__endian(full_type)
                kind = self.__kind(full_type,endian)

                # Get the comparaison test and result
                (test,result) = self.__test_result(part[2])
                
                # Get the value to check against
                data = self.__data(result)
                
                # Special characters
                mime = self.__mime(part[3:])

                # Calculate the size of the data to read in the file
                if kind == "string":
                    length = len(data)
                else:
                    length = self.data_size[kind]
                    
                # Append the line to the list
                acc_level.append(level)
                acc_direct.append(direct)
                acc_offset_type.append(offset_type)
                acc_offset_delta.append(offset_delta)
                acc_offset_relatif.append(offset_relatif)
                acc_endian.append(endian)
                acc_kind.append(kind)
                acc_oper.append(oper)
                acc_mask.append(mask)
                acc_test.append(test)
                acc_data.append(data)
                acc_length.append(length)
                acc_mime.append(mime)

                self.entries += 1

        f.close()
        
        self._level = tuple(acc_level)
        self._direct = tuple(acc_direct)
        self._offset_relatif = tuple(acc_offset_relatif)
        self._offset_type = tuple(acc_offset_type)
        self._offset_delta = tuple(acc_offset_delta)
        self._endian = tuple(acc_endian)
        self._kind = tuple(acc_kind)
        self._oper = tuple(acc_oper)
        self._mask = tuple(acc_mask)
        self._test = tuple(acc_test)
        self._data = tuple(acc_data)
        self._length = tuple(acc_length)
        self._mime = tuple(acc_mime)

    def write_cache (self,name):
        f = open (name,'wb')

        #import safe_pickle
        import cPickle as safe_pickle
        
        safe_pickle.dump(self._level,f)
        safe_pickle.dump(self._direct,f)
        safe_pickle.dump(self._offset_relatif,f)
        safe_pickle.dump(self._offset_type,f)
        safe_pickle.dump(self._offset_delta,f)
        safe_pickle.dump(self._endian,f)
        safe_pickle.dump(self._kind,f)
        safe_pickle.dump(self._oper,f)
        safe_pickle.dump(self._mask,f)
        safe_pickle.dump(self._test,f)
        safe_pickle.dump(self._data,f)
        safe_pickle.dump(self._length,f)
        safe_pickle.dump(self._mime,f)

        f.close()
    
    def read_cache (self,name):
        f = open (name,'rb')

        #import safe_pickle
        import cPickle as safe_pickle
        
        self._level = tuple(safe_pickle.load(f))
        self._direct = tuple(safe_pickle.load(f))
        self._offset_relatif = tuple(safe_pickle.load(f))
        self._offset_type = tuple(safe_pickle.load(f))
        self._offset_delta = tuple(safe_pickle.load(f))
        self._endian = tuple(safe_pickle.load(f))
        self._kind = tuple(safe_pickle.load(f))
        self._oper = tuple(safe_pickle.load(f))
        self._mask = tuple(safe_pickle.load(f))
        self._test = tuple(safe_pickle.load(f))
        self._data = tuple(safe_pickle.load(f))
        self._length = tuple(safe_pickle.load(f))
        self._mime = tuple(safe_pickle.load(f))

        self.entries = len(self._level)

        f.close()

    def classify(self,name):
        if not self.entries:
            raise StandardError, "Not initialised properly"
        
        in_level = 0
        wait_for = 0
        result = ""
        results = {}
        f = open(name,'rb')

        f.seek(0,2)
        file_length = f.tell()

        for i in range(self.entries):
            level = self._level[i]

            if in_level == 0 and level > 0:
                continue

            if wait_for != 0 and level > wait_for:
                continue

            direct = self._direct[i]
            offset_type = self._offset_type[i]
            offset_delta = self._offset_delta[i]
            offset_relatif = self._offset_relatif[i]
            endian = self._endian[i]
            kind = self._kind[i]
            oper = self._oper[i]
            mask = self._mask[i]
            test = self._test[i]
            data = self._data[i]
            leng = self._length[i]
            mime = self._mime[i]

            value = 0
            success = 0
            replace = None

            # Calculate final file offset to read at
            if direct == 1:
                if file_length < offset_delta:
                    continue
                offset = offset_delta

            else:
                if file_length < offset_relatif:
                    continue

                f.seek(offset_relatif)
                try:
                    if offset_type == 'l':
                        offset = convert.little4(f.read(4))
                    elif offset_type == 'L':
                        offset = convert.big4(f.read(4))
                    elif offset_type == 's':
                        offset = convert.little2(f.read(2))
                    elif offset_type == 'S':
                        offset = convert.big2(f.read(2))
                    elif offset_type == 'b' or offset_type == 'B':
                        offset = ord(f.read(1))

                    offset += offset_delta
                except:
                    continue

            # If it is out of the file then the test fails.
            if file_length < offset:
                continue

            f.seek(offset)

            try:
                extract=f.read(leng)
            except:
                raise IOError, "file.read() problem"

            # if we can not read what we need
            if len(extract) != leng:
                continue

            if in_level > 0:
                if level == 0:
                    # We performed a series of test and it is now completed
                    break
                elif level < in_level:
                    # We have another set of test to perform at lower level
                    in_level = level
                elif level > in_level+1:
                    # Fault in the file, can not go from ">" to ">>>" 
                    continue

            # Convert the data from the file if needed
            if kind == 'byte':
                value=ord(extract[0])
            elif kind == 'short':
                if endian == 'local':
                    value = convert.local2(extract)
                elif endian == 'little':
                    value = convert.little2(extract)
                elif endian == 'big':
                    value = convert.big2(extract)
                else:
                    raise StandardError, "Endian type unknown"
            elif kind == 'long':
                if endian == 'local':
                    value = convert.local4(extract)
                elif endian == 'little':
                    value = convert.little4(extract)
                elif endian == 'big':
                    value = convert.big4(extract)
                else:
                    raise StandardError, "Endian type unknown"
            elif kind == 'date':
                # XXX: Not done yet
                pass
                pass
            elif kind == 'ldate':
                # XXX: Not done yet
                pass
                pass
            elif kind == 'string':
                # Nothing to do
                pass
            elif kind == 'pstring':
                #Convert it to be like a string
                size=ord(extract[0])
                # Not sure the test is right (at one byte)
                if file_length < offset + size:
                    extract=f.read(size)
                    leng = size
                    kind = "string"
            else:
                raise StandardError, "Type " + str(kind) + " not recognised"

            if oper == '&':
                value &= mask
            elif oper == '^':
                value ^= mask
            elif oper == '':
                pass
            else:
                raise StandardError, "Binary operator unknown " + str(oper) 


            if test == '=':
                if kind == 'string':
                    if len(data) == len(extract):
                        success=1
                        for index in range(len(data)):
                            if ord(extract[index]) != data[index]:
                                success = 0
                else:
                    success = (data[0] == value)
                    replace = value
            elif test == '>':
                if kind == 'string':
                    if ord(extract[0]) != 0:
                        limit=0
                        replace = "" 
                        while 1:
                            if ord(extract[0]) == 0 or limit > 100:
                                break
                            replace += extract
                            extract = f.read(1)
                            limit += 1
                        if limit <= 100:
                            success = 1
                else:
                    success = (value > data[0])
                    replace = value
            elif test == '<':
                if kind == 'string':
                    success = 1 
                    
                    minimum = min(len(data),len(extract))
                    if len(extract) > minimum:
                        success = 0
                    else:
                        for index in range(minimum):
                            if data[index] > extract[index]:
                                success = 0
                                break
                else:
                    success = (value < data[0])
                    replace = value
            elif test == '&':
                # XXX: To be tested with a known file
                success = ((value & data[0]) == data[0])
                replace = value
            elif test == '^':
                # XXX: To be tested with a known file
                success = ((value ^ data[0]) == 0)
                replace = value
            elif test == '!':
                # XXX: To be tested with a known file
                # XXX: Wrong so must be a binary inversion test
                # success = (value != data[0])
                success = 0
                replace = value
            elif test == 'x':
                # XXX: copy from the code in test == '>', should create a function
                if kind == 'string':
                    limit=0
                    while 1:
                        if ord(extract[0]) == 0 or limit > 100:
                            break

                        if replace == None:
                            replace = extract
                        else:
                            replace += extract

                        extract = f.read(1)
                        limit += 1
                    if limit <= 100:
                        success = 1
                else:
                    success = 1
                    replace = value
            else:
                raise StandardError, "test used '"+test+"' is not defined"

            if success:
                if level > 0 and in_level == 0:
                    continue
                if in_level == 0 and level == 0:
                    in_level = 1
                elif in_level == (level-1):
                    in_level = level

                if replace is not None:
                    try:
                        mime = mime % replace
                    except:
                        pass
                if mime != []:
                    result += mime
                    result += ' '

                wait_for = 0
            else:
                if wait_for <= level:
                    wait_for = level 
                    

        f.close()

        if in_level == 0:
            return "unknow"

        return result.rstrip()

if __name__ == '__main__':
    import sys

    try:    binname = sys.argv[1]
    except:    binname = "neko.xpm"

    try:    filename = sys.argv[2]
    except:    filename = "magic.circle"

    try:    cachename = sys.argv[3]
    except:    cachename = "magic.cache"

    magic = Magic(filename,cachename)

    classify = magic.classify(binname)
    
    print binname + ": " + classify

# vim: expandtab:shiftwidth=4 :
