#!/usr/bin/env python


from gtk import *
from utility import Task
from error import Error
import check
import types

"""This doc comment contains a selection of random thoughts on
aspects of implementing whiteboard idea well.  Discuss ideas on #circle,
then have someone commit the conclusions to this file.

If you see something here that you can't see the purpose of, then
don't assume that there is one: add a comment querying it.  The
comments in this block are intended to be self-contained.


Currently the whiteboard attempts to send way too much traffic.

There are a number of ways to go re whiteboard:

i) Don't have it in circle; look at an existing whiteboard-providing
program, and use that by the side of one's circle client.

ii) Go an entirely generic route re drawing prog, and have circle send
diffs (xdelta) between versions of the produced file.

Vector-based drawing tools are quite good from a network traffic
perspective.

Apple's Newton PDA did cool things here: converting free-hand drawings
into lines, ovals, rectangles etc: guessing what shapes you actually
wanted to draw.  This acts as good compression while allowing quick
drawing.  If this is considered useful, then some people at Monash
University (Bernd, Kim, tonyj) may be able to provide help.


If using circle, then sending UDP packets without checking for errors
etc. is one interesting route to take.  (The intent being to reduce
latency.)


Colours
~~~~~~~

Suggest that each person's scribble be shown in different colour.

The intuitive-idea would be that there's a common whiteboard and each
person has their own marker.

(Can a key/legend somewhere showing the mapping between colour and nick.)

There's benefit in a given person being shown in the same colour to
all viewers: people will probably start talking about ``the red
squiggle'', which will be a problem if it isn't red for everyone.

One possibility is to form colour as a function of the person's
identity (circle-person:... URI).  As a simple example, take Hue from
identity, and keep Value and Saturation at constant levels.  Try to
pick the function such that each colour in its range is easily visible
on a white background.  As a starting point, jiri suggests Value ~=
45-50%; Saturation ~= 60%.

(These figures can easily be changed later, even if different client
versions have different values: if they are constant between all the
colours that a given person sees then that person won't say "the grey
squiggle" if each are equally grey.)

An advantage of using a person's identity is that Fred always has the
same colour, and one associates the colour with the person.

(One minor other pleasing property is that colour doesn't need to be
sent over the network: clients already have all the information they
need.  This slightly simplifies the network protocol and reduces
bandwidth.)

There can be problems in conversations between people whose identities
are such that they're assigned very similar colours.

One solution is to perform small colour shifts when this occurs; just
enough to maintain some minimum distance between colours.  (This is
algorithmically fairly simple, btw, assuming the input to the
algorithm is just a list of identities.)

This minimum distance must of course be partly a function of the
number of participants; e.g. max(10%, 100% / n_participants).


Special Case Diagrams
~~~~~~~~~~~~~~~~~~~~~

Graphs (as in nodes & arcs) may be useful special case, especially if
auto-relayout is available when adding a node.  Even just having the
arrow ends kept with their nodes is a useful saving.

Possibly this should be done with dia, and circle sending file diffs.
(Dia already supports python plugins.)


Editing text files may be another special case.  It is a more common
example than editing graphs, though perhaps it doesn't have as much
benefit over existing available techniques (e.g. CVS) as graph
drawing.

Conflicts are a problem to some extent with any shared document,
though conflicts are clearly a much bigger problem for text files
than with simple squiggle-on-whiteboard.



"""

class Whiteboard:

    def __init__(self,chat):
        self.ink = 0
        self.width=300
        self.height=300
        self.chat=chat
        self.points = [ ]

        self.window = GtkWindow(WINDOW_TOPLEVEL)
        self.area = GtkDrawingArea()
        self.area.add_events(GDK.BUTTON_RELEASE_MASK)
        self.area.add_events(GDK.BUTTON_PRESS_MASK)
        self.area.add_events(GDK.POINTER_MOTION_MASK)
        self.area.size(self.width, self.height)
        self.window.add(self.area)
        self.area.realize()
        win = self.area.get_window()
        self.pixmap = create_pixmap(win,self.width, self.height, -1)
        self.context = win.new_gc()
        self.cmap = self.area.get_colormap()
        self.black = self.cmap.alloc(0, 0, 0)
        self.white = self.cmap.alloc(65535, 65535, 65535)

        self.clear()
        self.area.connect("expose-event", self.pix_expose)
        self.area.connect("motion_notify_event", self.on_move)
        self.area.connect("button_release_event", self.on_release)
        self.area.connect("button_press_event", self.on_press)
        self.window.connect("destroy", self.on_close)
        self.window.show_all()

    def clear(self):
        self.context.foreground = self.white
        draw_rectangle(self.pixmap, self.context, 1,0, 0,self.width,self.height)
        self.context.foreground = self.black
        self.area.draw_pixmap(self.area.get_style().fg_gc[0], self.pixmap, 0, 0, 0, 0, -1, -1)
            
    def pix_expose(self,area, event):
        area.draw_pixmap(area.get_style().fg_gc[0], self.pixmap, 0, 0, 0, 0, -1, -1)

    def on_close(self,win):
        self.chat.whiteboard=None

    def on_move(self,area,_e):
        if self.ink:
            draw_line(self.pixmap,self.context,_e.x, _e.y,self.oldx,self.oldy)
            self.oldx=_e.x
            self.oldy=_e.y
            self.points.append('%d %d'%(_e.x,_e.y))
            area.draw_pixmap(area.get_style().fg_gc[0],self.pixmap, 0, 0, 0, 0, -1, -1)

    def on_press(self,area,_e):
        if _e.button == 1:
            self.ink = 1
            draw_point(self.pixmap,self.context,_e.x, _e.y)
            self.oldx=_e.x
            self.oldy=_e.y
            self.points = [ ]
            self.points.append('%d %d'%(_e.x,_e.y))
            area.draw_pixmap(area.get_style().fg_gc[0], self.pixmap, 0, 0, 0, 0, -1, -1)
        elif _e.button == 3:
            self.menu(area,_e)

    def on_release(self,area,_e):
        if _e.button <> 1:
            return
        self.ink = 0
        area.draw_pixmap(area.get_style().fg_gc[0], self.pixmap, 0, 0, 0, 0, -1, -1)

        #note : use compression (eg bezier, or gzip)

        Task(send_stroke, self.chat, self.points).start()

    def draw_string(self,request):
        check.check_matches(request, (types.StringType, 'any'))
        
        if request[1]=='clear':
            self.clear()
        elif check.matches(request[1], [types.StringType]):
            old_x=None
            old_y=None
            for item in request[1]:
                ok = 0
                try:
                    x_str,y_str = item.split()
                    x = int(x_str)
                    y = int(y_str)
                    ok = 1
                except:
                    print "Bad draw coord string `%s'." % item
                if not ok:
                    continue  # Or break?
                if old_x is None:
                    draw_point(self.pixmap,self.context,x,y)
                else:
                    draw_line(self.pixmap,self.context,x,y,old_x,old_y)
                old_x=x
                old_y=y
            self.area.draw_pixmap(self.area.get_style().fg_gc[0], self.pixmap, 0, 0, 0, 0, -1, -1)
        else:
            print "Unrecognized draw request `%s'." % str(request[1])


    def menu(self,w,ev):
        menu = []
        def clear(d=w,self=self):
            self.clear()
        menu.append(("clear",clear))

        def clear_all(d=w,self=self):
            self.points = 'clear'
            Task(send_stroke, self.chat, self.points).start()
            self.clear()
            
        menu.append(("clear all",clear_all))

        menuwidget = GtkMenu()
        for (text, action) in menu:
            mi = GtkMenuItem(text)
            mi.connect("activate", action)
            mi.show()
            menuwidget.append(mi)
        menuwidget.popup(None,None,None,ev.button,ev.time)



def send_stroke(chat, points):
    return  # see FIXME comment below.
    dests = chat.channel[:]

    dest_list = [ ]
    dest_address  = [ ]
    dest_names = [ ]
    dest_bad_names = [ ]
    recipient_list = [ ]
    channel_pipes = { }
            
    for dest in dests:
        if dest[0]=='#':
            channel_pipes[dest] = chat.app.channels.sub_list_pipe(dest)
            recipient_list.append((None,None,dest))
            continue
        acq = chat.app.name_server.locate(dest)
        acq.start_watching(chat.node,1)

        acq.lock.acquire()
        online = acq.online
        address = acq.address
        name = acq.name
        username = acq.info['name']
        acq.lock.release()
             
        if online:
            dest_list.append(acq)
            dest_address.append(address)
            dest_names.append(dest)
            recipient_list.append((name,address,username))
                
        my_details = (
            chat.app.name_server.public_key_name,
            chat.app.name_server.name
            )

    #sig = chat.app.name_server.sign(points)
    #message = ('chat whiteboard',my_details,recipient_list,message,sig)
    message = ('chat whiteboard', points)

    # FIXME: dest_list is a list of Acquaintance; node.call wants an
    # address.  See chat.py:chat_tell_task.
    # However, I suggest that we don't want to be sending a message
    # (to each of n recipients) each time the mouse moves.
    for addr in dest_list:
        # Note: 'chat whiteboard' not present in 0.32 and earlier,
        # and not present for text interface.
        try:
            chat.node.call(addr, message)
        except Error, e:
            if e.message == 'Request not understood: chat whiteboard':
                # TODO: Handle this more gracefully.
                #incapable_clients[addr] = 1
                print 'Client at address ' + str(addr) + ' has no whiteboard.'
            else:
                raise e


# vim: expandtab:shiftwidth=4 :
